@@ -28,7 +28,7 @@
  *
  */
 
-static const char revision[] = "$Id: qslog.c,v 1.106 2017/06/29 05:24:33 pbuchbinder Exp $";
+static const char revision[] = "$Id: qslog.c,v 1.114 2017-10-02 15:44:03 pbuchbinder Exp $";
 
 #include <stdio.h>
 #include <string.h>
@@ -45,6 +45,8 @@ static const char revision[] = "$Id: qslog.c,v 1.106 2017/06/29 05:24:33 pbuchbi
 #include <regex.h>
 #include <time.h>
 
+#include <pcre.h>
+
 /* apr */
 #include <apr.h>
 #include <apr_portable.h>
@@ -67,11 +69,25 @@ static const char revision[] = "$Id: qslog.c,v 1.106 2017/06/29 05:24:33 pbuchbi
 #define QS_GENERATIONS 14
 #define EVENT_DELIM ','
 #define QSEVENTPATH "QSEVENTPATH" /* varibale name to find event definitions */
+#define QSCOUNTERPATH "QSCOUNTERPATH" /* counter rule definitions */
+#define COUNTER_PATTERN "([a-zA-Z0-9_]+):([a-zA-Z0-9_]+)[-]([0-9]+)[*]([a-zA-Z0-9_]+)[/]([0-9]+)=([0-9]+)"
 
 /* ----------------------------------
  * structures
  * ---------------------------------- */
 
+typedef struct {
+  const char *name;
+  int limit;
+  int count;
+  int total;
+  time_t start;
+  int duration;
+  const char *inc;
+  const char *dec;
+  int decVal;
+} counter_rec_t;
+
 typedef struct {
   long request_count;
   long status_1;
@@ -102,7 +118,9 @@ typedef struct {
   long status_5;
   long status_304;
   long connections;
+  unsigned long max;
   apr_table_t *events;
+  apr_table_t *counters;
   apr_pool_t *pool;
   long get;
   long post;
@@ -179,8 +197,9 @@ static stat_rec_t* m_stat_sub = NULL;
 static time_t m_qs_expiration = 60 * 10;
 
 static apr_table_t *m_ip_list[NUM_EVENT_TABLES];   /* IP session store */
+static int m_ip_log_max = 0;                       /* already reached store limit */
 static apr_table_t *m_user_list[NUM_EVENT_TABLES]; /* user session store */
-static int m_log_max = 0;                          /* already reached store limit */
+static int m_usr_log_max = 0;                      /* already reached store limit */
 static int m_hasGC = 0;                            /* sep. gc thread or not */
 static qs_event_t **m_gc_event_list = NULL;        /* list of entries to delete */
 
@@ -283,10 +302,10 @@ static int qs_tableSelector(const char *str) {
  *
  * @param l_qs_event Pointer to the event list.
  * @param id Identifer, e.g. IP address or user tracking cookie
- *
+ * @param type which counter is used (either 'I' or 'U')
  * @return event counter (number of updates) for the provided id
  */
-static long qs_insertEventT(apr_table_t **list0, const char *id) {
+static long qs_insertEventT(apr_table_t **list0, const char *id, const char *type) {
   qs_event_t *lp;
   int select = qs_tableSelector(id);
   apr_table_t *list = list0[select];
@@ -298,14 +317,20 @@ static long qs_insertEventT(apr_table_t **list0, const char *id) {
     return lp->count;
   }
   if(apr_table_elts(list)->nelts >= MAX_EVENT_ENTRIES) {
-    if(m_log_max == 0) {
+    if((type[0] == 'I' && m_ip_log_max == 0) ||
+       (type[0] == 'U' && m_usr_log_max == 0)) {
       char time_string[1024];
       time_t tm = time(NULL);
       struct tm *ptr = localtime(&tm);
       strftime(time_string, sizeof(time_string), "%a %b %d %H:%M:%S %Y", ptr);
-      fprintf(stderr, "[%s] [notice] qslog: reached event count limit\n",
-              time_string);
-      m_log_max = 1;
+      fprintf(stderr, "[%s] [notice] qslog: reached event (%s) count limit\n",
+              time_string, type);
+      if(type[0] == 'I') {
+        m_ip_log_max = 1;
+      }
+      if(type[0] == 'U') {
+        m_usr_log_max = 1;
+      }
     }
     return 0;
   }
@@ -672,6 +697,9 @@ static void printStat2File(FILE *f, char *timeStr, stat_rec_t *stat_rec,
   avms[0] = '\0';
   custom[0] = '\0';
 
+  m_ip_log_max = 0;
+  m_usr_log_max = 0;
+
   if(stat_rec->i_byte_count != -1) {
     sprintf(bis, NBIS";%lld;", stat_rec->i_byte_count/LOG_INTERVAL);
   }
@@ -832,6 +860,64 @@ static void printStat2File(FILE *f, char *timeStr, stat_rec_t *stat_rec,
   fprintf(f, "\n");
 }
 
+/**
+ * updates the counter by event or status conditions
+ */
+static void qs_updateCounter(apr_pool_t *pool, char *E, char *S, apr_table_t *counters) {
+  time_t ltime;
+  apr_table_entry_t *entry;
+  int i;
+  if(counters == NULL) {
+    return;
+  }
+  if(S == 0 && E == NULL) {
+    return;
+  }
+  qs_time(&ltime);
+  entry = (apr_table_entry_t *) apr_table_elts(counters)->elts;
+  for(i = 0; i < apr_table_elts(counters)->nelts; i++) {
+    counter_rec_t *c = (counter_rec_t *)entry[i].val;
+    if(c->start && ((c->start + c->duration) < ltime)) {
+      // expired
+      c->start = 0;
+      c->count = 0;
+    }
+  }
+  for(i = 0; i < apr_table_elts(counters)->nelts; i++) {
+    counter_rec_t *c = (counter_rec_t *)entry[i].val;
+    if(S) {
+      if((strncmp(c->name, "STATUS", 6) == 0) &&
+         (strstr(c->inc, S) != NULL)) {
+        if(c->start == 0) {
+          c->start = ltime;
+        }
+        c->count++;
+        if(c->count == c->limit) {
+          c->total++;
+        }
+      }
+    }
+    if(E) {
+      if(strstr(c->inc, E)) {
+        if(c->start == 0) {
+          c->start = ltime;
+        }
+        c->count++;
+        if(strstr(c->dec, E)) {
+          if(c->count > c->decVal) {
+            c->count = c->count - c->decVal;
+          } else {
+            c->count = 0;
+          }
+        }
+        if(c->count == c->limit) {
+          c->total++;
+        }
+      }
+    }
+  }
+}
+
 static void qs_updateEvents(apr_pool_t *pool, char *E, apr_table_t *events) {
   if(!E[0]) {
     return;
@@ -865,10 +951,55 @@ static void qs_updateEvents(apr_pool_t *pool, char *E, apr_table_t *events) {
   }
 }
 
+/**
+ * Reads the counter rule file, each line contains:
+ * <name>:<event>-<n>*<event>/<duration>=<limit>
+ */
+static void qsInitCounter(apr_pool_t *pool, apr_table_t *counters) {
+  const char *envFile = getenv(QSCOUNTERPATH);
+  if(envFile != NULL) {
+    const char *errptr = NULL;
+    int erroffset;
+    int ovector[100];
+    pcre *pcrestat = pcre_compile(COUNTER_PATTERN, PCRE_CASELESS, &errptr, &erroffset, NULL);
+    FILE *file = fopen(envFile, "r"); 
+    if(file != NULL) {
+      char line[MAX_LINE];
+      while(!qs_getLinef(line, sizeof(line), file)) {
+        if(pcre_exec(pcrestat, NULL, line, strlen(line), 0, 0, ovector, 100) >= 0) {
+          counter_rec_t *c = apr_pcalloc(pool, sizeof(counter_rec_t));
+          line[ovector[2] + ovector[3] - ovector[2]] = '\0';
+          line[ovector[4] + ovector[5] - ovector[4]] = '\0';
+          line[ovector[6] + ovector[7] - ovector[6]] = '\0';
+          line[ovector[8] + ovector[9] - ovector[8]] = '\0';
+          line[ovector[10] + ovector[11] - ovector[10]] = '\0';
+ 
+          c->name = apr_pstrdup(pool, &line[ovector[2]]);
+          c->count = 0;
+          c->total = 0;
+          c->start = 0;
+          c->inc = apr_pstrdup(pool, &line[ovector[4]]);
+          c->decVal = atoi(&line[ovector[6]]);
+          c->dec = apr_pstrdup(pool, &line[ovector[8]]);
+          c->duration = atoi(&line[ovector[10]]);
+          c->limit = atoi(&line[ovector[12]]);
+          if(m_verbose) {
+            fprintf(stderr, "%s : %s - (%d * %s) / %d = %d\n",
+                    c->name, c->inc, c->decVal, c->dec, c->duration, c->limit);
+          }
+          apr_table_setn(counters, c->name, (char *)c);
+        }
+      }
+      fclose(file);
+    }
+  }
+}
+
 /**
  * Initializes the event table by the events specified within the
  * file whose path is defined by the QSEVENTPATH environment
  * variable.
+ * File contains event names, separated by comma and/or new line.
  *
  * @param pool To allocate memory
  * @param events Table to init
@@ -1100,7 +1231,7 @@ static void updateUrl(apr_pool_t *pool, char *R, char *S, long tmems) {
  */
 static void updateClient(apr_pool_t *pool, char *T, char *t, char *D, char *S,
                          char *BI, char *B, char *R, char *I, char *U, char *Q,
-                         char *E, char *k, char *C, char *ct, long tme, long tmems,
+                         char *E, char *k, char *C, char *M, char *ct, long tme, long tmems,
                          char *m) {
   client_rec_t *client_rec;
   const char *id = I; // ip
@@ -1144,7 +1275,9 @@ static void updateClient(apr_pool_t *pool, char *T, char *t, char *D, char *S,
     client_rec->status_5 = 0;
     client_rec->status_304 = 0;
     client_rec->connections = 0;
+    client_rec->max = 0;
     client_rec->events = apr_table_make(pool, 100);
+    client_rec->counters = apr_table_make(pool, 10);
     client_rec->pool = pool;
     client_rec->get = 0;
     client_rec->post = 0;
@@ -1156,6 +1289,7 @@ static void updateClient(apr_pool_t *pool, char *T, char *t, char *D, char *S,
     client_rec->end_s = client_rec->start_s + 1; // +1 prevents div by 0
     client_rec->firstLine = m_lines;
     qsInitEvent(pool, client_rec->events);
+    qsInitCounter(pool, client_rec->counters);
     apr_table_setn(m_client_entries, tid, (char *)client_rec);
   } else {
     qs_time(&client_rec->end_s);
@@ -1200,6 +1334,12 @@ static void updateClient(apr_pool_t *pool, char *T, char *t, char *D, char *S,
       client_rec->other++;
     }
   }
+  if(M && M[0]) {
+    long max = atol(M);
+    if(max > client_rec->max) {
+      client_rec->max = max;
+    }
+  }
   if(m) {
     if(strcasecmp(m, "get") == 0) {
       client_rec->get++;
@@ -1231,6 +1371,7 @@ static void updateClient(apr_pool_t *pool, char *T, char *t, char *D, char *S,
   if(E != NULL) {
     qs_updateEvents(client_rec->pool, E, client_rec->events);
   }
+  qs_updateCounter(client_rec->pool, E, S, client_rec->counters);
   return;
 }
 
@@ -1278,11 +1419,11 @@ static void updateRec(stat_rec_t *rec, char *T, char *t, char *D, char *S,
   }
   if(I != NULL) {
     /* update/store client IP */
-    qs_insertEventT(m_ip_list, I);
+    qs_insertEventT(m_ip_list, I, "I");
   }
   if(U != NULL) {
     /* update/store user */
-    qs_insertEventT(m_user_list, U);
+    qs_insertEventT(m_user_list, U, "U");
   }
   if(B != NULL) {
     /* transferred bytes */
@@ -1376,6 +1517,7 @@ static void updateStat(apr_pool_t *pool, const char *cstr, char *line) {
   char *s = NULL; /* sum */
   char *a = NULL; /* avarage 1 */
   char *A = NULL; /* average 2 */
+  char *M = NULL; /* max */
   char *E = NULL; /* events */
   char *ct = NULL; /* content type */
   char *m = NULL; /* method */
@@ -1461,6 +1603,10 @@ static void updateStat(apr_pool_t *pool, const char *cstr, char *line) {
       if(l != NULL && l[0] != '\0') {
         A = cutNext(&l);
       }
+    } else if(c[0] == 'M') {
+      if(l != NULL && l[0] != '\0') {
+        M = cutNext(&l);
+      }
     } else if(c[0] == 'E') {
       if(l != NULL && l[0] != '\0') {
         E = cutNext(&l);
@@ -1506,6 +1652,10 @@ static void updateStat(apr_pool_t *pool, const char *cstr, char *line) {
     stripNum(&A);
     qsNoFloat(A);
   }
+  if(M != NULL) {
+    stripNum(&M);
+    qsNoFloat(M);
+  }
   tme = 0;
   if(T != NULL || t != NULL || D != NULL) {
     /* response duration */
@@ -1525,7 +1675,7 @@ static void updateStat(apr_pool_t *pool, const char *cstr, char *line) {
     }
   }
   if(m_offline_count) {
-    updateClient(pool, T, t, D, S, BI, B, R, I, U, Q, E, k, C, ct, tme, tmems, m);
+    updateClient(pool, T, t, D, S, BI, B, R, I, U, Q, E, k, C, M, ct, tme, tmems, m);
   } else if(m_offline_url) {
     if((tmems) == 0 && (tme > 0)) {
       tmems = 1000 * tme;
@@ -1831,19 +1981,21 @@ static void usage(const char *cmd, int man) {
   }
   qs_man_print(man, "%s - collects request statistics from access log data.\n", cmd);
   printf("\n");
+
   if(man) {
     printf(".SH SYNOPSIS\n");
   }
   qs_man_print(man, "%s%s -f <format_string> -o <out_file> [-p[c|u[c]] [-v]] [-x [<num>]] [-u <name>] [-m] [-c <path>]\n", man ? "" : "Usage: ", cmd);
   printf("\n");
+
   if(man) {
     printf(".SH DESCRIPTION\n");
   } else {
     printf("Summary\n");
   }
-  qs_man_print(man, "%s is a real time access log analyzer. It collects\n", cmd);
-  qs_man_print(man, "the data from stdin. The output is written to the specified\n");
-  qs_man_println(man, "file every minute and includes the following entries:\n");
+  qs_man_print(man, "%s is a real time access log analyzer. It collects the data from stdin.\n", cmd);
+  qs_man_print(man, "The output is written to the specified file every minute and includes the\n");
+  qs_man_println(man, "following entries:\n");
   qs_man_println(man, "  - requests per second ("NRS")\n");
   qs_man_println(man, "  - number of requests within measured time (req)\n");
   qs_man_println(man, "  - bytes sent to the client per second ("NBS")\n");
@@ -1865,6 +2017,7 @@ static void usage(const char *cmd, int man) {
   qs_man_print(man, "    keep-alive, qL=request/response slow down, qs=serialized request, \n");
   qs_man_print(man, "    qA=connection abort, qU=new user tracking cookie)\n");
   printf("\n");
+
   if(man) {
     printf(".SH OPTIONS\n");
   } else {
@@ -1894,6 +2047,7 @@ static void usage(const char *cmd, int man) {
   qs_man_println(man, "     s arbitrary counter to add up (sum within a minute)\n");
   qs_man_println(man, "     a arbitrary counter to build an average from (average per request)\n");
   qs_man_println(man, "     A arbitrary counter to build an average from (average per request)\n");
+  qs_man_println(man, "     M maximum of an ever reached value, available in -pc mode only\n");
   qs_man_println(man, "     E comma separated list of event strings\n");
   qs_man_println(man, "     c content type (%%{content-type}o), available in -pc mode only\n");
   qs_man_println(man, "     m request method (GET/POST) (%%m), available in -pc mode only\n");
@@ -1912,7 +2066,7 @@ static void usage(const char *cmd, int man) {
   qs_man_print(man, "     The option \"-pc\" may be used alternatively if you want to gather request\n");
   qs_man_print(man, "     information per client (identified by IP address (I) or user tracking id (U)\n");
   qs_man_print(man, "     showing how many request each client has performed within the captured period\n");
-  qs_man_print(man, "     of time). \"-pc\" supports the format characters IURSBTtDkEcm.\n");
+  qs_man_print(man, "     of time). \"-pc\" supports the format characters IURSBTtDkMEcm.\n");
   qs_man_print(man, "     The option \"-pu\" collects statistics on a per URL level (supports format\n");
   qs_man_print(man, "     characters RSTtD).\n");
   qs_man_print(man, "     \"-puc\" is very similar to \"-pu\" but cuts the end (handler) of each URL.\n");
@@ -1941,6 +2095,50 @@ static void usage(const char *cmd, int man) {
   qs_man_print(man, "     identifier and a regular expression to identify a request seprarated\n");
   qs_man_print(man, "     by a colon, e.g., 01:^(/a)|(/c). The regular expressions are matched against\n");
   qs_man_print(man, "     the log data element which has been identified by the 'C' format character.\n");
+
+  printf("\n");
+  if(man) {
+    printf(".SH VARIABLES\n");
+  } else {
+    printf("Variables\n");
+  }
+  qs_man_print(man, "The following environment variables are known to %s:\n", cmd);
+  if(man) printf("\n");
+  if(man) printf(".TP\n");
+  qs_man_print(man, "  "QSEVENTPATH"=<path>\n");
+  if(man) printf("\n");
+  qs_man_print(man, "     Defines a file containing a comma or new line separated list\n");
+  qs_man_print(man, "     of known event strings expected within the log filed identified\n");
+  qs_man_print(man, "     by the 'E' format character.\n");
+  if(man) printf("\n.TP\n");
+  qs_man_print(man, "  "QSCOUNTERPATH"=<path>\n");
+  if(man) printf("\n");
+  qs_man_print(man, "     Defines a file containing a by new line separated list of rules which\n");
+  qs_man_print(man, "     reflect possible QS_ClientEventLimitCount directive settings (for\n");
+  qs_man_print(man, "     simulation purpose / -pc option). The 'E' format character defines the event\n");
+  qs_man_print(man, "     string in the log to match (literal string) the 'event1' and 'event2' event\n");
+  qs_man_print(man, "     names against.\n");
+  printf("\n");
+  if(man) printf("\n");
+  qs_man_print(man, "     Rule syntax: <name>:<event1>-<n>*<event2>/<duration>=<limit>\n");
+  printf("\n");
+  qs_man_println(man, "     'name' defines the name you have given to the rule entry and is logged along with\n");
+  qs_man_print(man, "      with the number of times the 'limit' has been reached within the 'duration'.\n");
+  if(man) printf("\n");
+  qs_man_println(man, "     'event1' defines the variable name (if found in 'E') to increment the counter.\n");
+  qs_man_println(man, "     'event2' defines the variable name (if found in 'E') to decrement the counter (and\n");
+  qs_man_print(man, "      the parameter 'n' defines by how much).\n");
+  if(man) printf("\n");
+  qs_man_println(man, "     'duration' defines the measure interval (in seconds) used for the\n");
+  qs_man_print(man, "      QS_ClientEventLimitCount directive.\n");
+  if(man) printf("\n");
+  qs_man_println(man, "     'limit' defines the threshold (number) defined for the QS_ClientEventLimitCount\n");
+  qs_man_print(man, "      directive.\n");
+  if(man) printf("\n");
+  printf("\n");
+  qs_man_print(man, "     Note: If the 'name' parameter is prefixed by 'STATUS', the rule is applied against\n");
+  qs_man_print(man, "     the HTTP status code 'S' and the 'event1' string shall contain a list of relevant\n");
+  qs_man_print(man, "     status codes separated by an underscore (while 'event2' is ignored).\n");
   printf("\n");
   if(man) {
     printf(".SH EXAMPLE\n");
@@ -2074,7 +2272,7 @@ int main(int argc, const char *const argv[]) {
         if(strchr(config, 'm')) {
           m_methods = 1;
         }
-        if(strchr(config, 's') || strchr(config, 'a') || strchr(config, 'A')) {
+        if(strchr(config, 's') || strchr(config, 'a') || strchr(config, 'A') || strchr(config, 'M')) {
           // enable custom counter
           m_customcounter = 1;
         }
@@ -2161,12 +2359,6 @@ int main(int argc, const char *const argv[]) {
   if(m_offline_url) {
     int i;
     apr_table_entry_t *entry;
-    long request_count = 0;
-    long request_count_min = -1;
-    long request_count_max = 0;
-    long long duration_count_ms = 0;
-    long long duration_count_ms_min = -1;
-    long long duration_count_ms_max = 0;
     m_url_entries = apr_table_make(pool, MAX_CLIENT_ENTRIES + 1);
     if(config == NULL) usage(cmd, 0);
     readStdinOffline(pool, config);
@@ -2182,26 +2374,6 @@ int main(int argc, const char *const argv[]) {
     entry = (apr_table_entry_t *) apr_table_elts(m_url_entries)->elts;
     for(i = 0; i < apr_table_elts(m_url_entries)->nelts; i++) {
       url_rec_t *url_rec = (url_rec_t *)entry[i].val;
-      request_count += url_rec->request_count;
-      if(request_count_min == -1) {
-        request_count_min = url_rec->request_count;
-      }
-      if(request_count_min > url_rec->request_count) {
-        request_count_min = url_rec->request_count;
-      }
-      if(request_count_max < url_rec->request_count) {
-        request_count_max = url_rec->request_count;
-      }
-      duration_count_ms += url_rec->duration_count_ms;
-      if(duration_count_ms_min == -1) {
-        duration_count_ms_min = url_rec->duration_count_ms;
-      }
-      if(duration_count_ms_min > url_rec->duration_count_ms) {
-        duration_count_ms_min = url_rec->duration_count_ms;
-      }
-      if(duration_count_ms_max < url_rec->duration_count_ms) {
-        duration_count_ms_max = url_rec->duration_count_ms;
-      }
       fprintf(m_f, "req;%ld;"
               "1xx;%ld;2xx;%ld;3xx;%ld;4xx;%ld;5xx;%ld;"
               NAVMS";%lld;%s\n",
@@ -2215,21 +2387,6 @@ int main(int argc, const char *const argv[]) {
               entry[i].key);
       
     }
-    fprintf(m_f, "req;%ld;"
-            ";;;;;;;;;;"
-            NAVMS";%lld;min;\n",
-            request_count_min,
-            duration_count_ms_min);
-    fprintf(m_f, "req;%ld;"
-            ";;;;;;;;;;"
-            NAVMS";%lld;max;\n",
-            request_count_max,
-            duration_count_ms_max);
-    fprintf(m_f, "req;%ld;"
-            ";;;;;;;;;;"
-            NAVMS";%lld;average;\n",
-            apr_table_elts(m_url_entries)->nelts ? (request_count / apr_table_elts(m_url_entries)->nelts) : 0,
-            request_count ? (duration_count_ms / request_count) : 0);
     if(file && m_f != stdout) {
       fclose(m_f);
     }
@@ -2315,6 +2472,18 @@ int main(int argc, const char *const argv[]) {
               esco,
               m,
               coverage);
+      if(m_customcounter) {
+        fprintf(m_f, "M;%ld;",
+                client_rec->max);
+      }
+      if(client_rec->counters) {
+        int c;
+        apr_table_entry_t *centry = (apr_table_entry_t *) apr_table_elts(client_rec->counters)->elts;
+        for(c = 0; c < apr_table_elts(client_rec->counters)->nelts; c++) {
+          counter_rec_t *cr = (counter_rec_t *)centry[c].val;
+          fprintf(m_f, "%s;%d;", cr->name, cr->total);
+        }
+      }
       if(m_ct) {
         fprintf(m_f, "html;%ld;css/js;%ld;img;%ld;other;%ld;",
                 client_rec->html,
